"""Model builder module.

Provides some helper methods for easier work with basyx sdk data model
"""

import json
import logging
import uuid
from pathlib import Path
from typing import Any

from basyx.aas import model

from aas_http_client.utilities import sdk_tools

_logger = logging.getLogger(__name__)


def create_unique_short_id(id_short: str) -> str:
    """Generate a unique identifier string by appending a UUID to the provided ID short.

    :param id_short: provided ID short
    :return: unique identifier
    """
    return f"{id_short}_{str(uuid.uuid4()).replace('-', '_')}"


def create_submodel_from_file(json_file: str = "") -> model.Submodel:
    """Creates a submodel from a JSON file file.

    :param id: ID of the submodel. If empty, the template's ID is used

    """
    # Define the path to the submodel template file
    sm_template_file = Path(json_file).resolve()

    # Check if the template file exists
    if not sm_template_file.exists():
        raise FileNotFoundError(f"Submodel template file not found: {sm_template_file}")

    submodel_data = {}
    try:
        # Load the template JSON file
        with Path.open(sm_template_file, "r", encoding="utf-8") as f:
            submodel_data = json.load(f)

        # Load the template JSON into a Submodel object
        submodel = sdk_tools.convert_to_object(submodel_data)

        # Ensure the loaded template is indeed a Submodel
        if not isinstance(submodel, model.Submodel):
            raise TypeError("Loaded template JSON structure is not a Submodel.")

    except Exception as e:
        _logger.error(f"Error loading submodel template: {e}")
        raise e

    return submodel


def create_base_submodel_element_property(
    id_short: str | None, type: model.datatypes, value: Any, display_name: str = "", description: str = ""
) -> model.Property:
    """Create a basic SubmodelElement of type Property."""
    sme = model.Property(id_short=id_short, value_type=type, value=value)

    if description:
        description_text = {"en": f"{description}"}
        sme.description = model.MultiLanguageTextType(description_text)

    if display_name:
        display_name_text = {"en": f"{display_name}"}
        sme.display_name = model.MultiLanguageNameType(display_name_text)

    return sme


def create_base_submodel_element_collection(
    id_short: str, value: list[model.SubmodelElement], display_name: str = "", description: str = ""
) -> model.SubmodelElementCollection:
    """Create a basic SubmodelElement of type SubmodelElementCollection."""
    sme = model.SubmodelElementCollection(id_short=id_short, value=value)

    if description:
        description_text = {"en": f"{description}"}
        sme.description = model.MultiLanguageTextType(description_text)

    if display_name:
        display_name_text = {"en": f"{display_name}"}
        sme.display_name = model.MultiLanguageNameType(display_name_text)

    return sme


def create_base_submodel(identifier: str, id_short: str, display_name: str = "", description: str = "") -> model.Submodel:
    """Create a basic Submodel.

    :param identifier: identifier of the Submodel
    :param id_short: ID short of the Submodel
    :param display_name: display name of the Submodel, defaults to ""
    :param description: description of the Submodel, defaults to ""
    :return: Submodel instance
    """
    sm = model.Submodel(identifier)
    sm.id_short = id_short

    if description:
        description_text = {"en": f"{description}"}
        sm.description = model.MultiLanguageTextType(description_text)

    if display_name:
        display_name_text = {"en": f"{display_name}"}
        sm.display_name = model.MultiLanguageNameType(display_name_text)

    return sm


def create_base_aas(
    identifier: str, id_short: str, global_asset_identifier: str, display_name: str = "", description: str = ""
) -> model.AssetAdministrationShell:
    """Create a basic AAS.

    :param identifier: identifier of the AAS
    :param id_short: ID short of the AAS
    :param global_asset_identifier: identifier of the global Asset
    :param display_name: display name of the AAS, defaults to ""
    :param description: description of the AAS, defaults to ""
    :return: AssetAdministrationShell instance
    """
    if not global_asset_identifier:
        global_asset_identifier = identifier

    asset_info = create_base_asset_information(global_asset_identifier)

    aas = model.AssetAdministrationShell(id_=identifier, asset_information=asset_info)
    aas.id_short = id_short

    if description:
        description_text = {"en": f"{description}"}
        aas.description = model.MultiLanguageTextType(description_text)

    if display_name:
        display_name_text = {"en": f"{display_name}"}
        aas.display_name = model.MultiLanguageNameType(display_name_text)

    return aas


def create_base_asset_information(identifier: str) -> model.AssetInformation:
    """Return a basic AssetInformation instance.

    :param id_short: short ID of the AssetInformation
    :param namespace: namespace of the AssetInformation, defaults to "basyx_python_aas_server"
    :return: AssetInformation instance
    """
    return model.AssetInformation(model.AssetKind.INSTANCE, identifier)


def create_reference(id: str) -> model.ModelReference:
    """Create a ModelReference.

    :param id: ID of the Submodel to reference
    :return: ModelReference instance
    """
    return model.ModelReference.from_referable(model.Submodel(id))



"""Implementation of Submodel related API calls."""

import json
import logging
from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel

if TYPE_CHECKING:
    from aas_http_client.classes.client.aas_client import AasHttpClient

from aas_http_client.utilities.encoder import encode_base_64
from aas_http_client.utilities.http_helper import (
    STATUS_CODE_200,
    STATUS_CODE_201,
    STATUS_CODE_204,
    STATUS_CODE_404,
    log_response,
)

_logger = logging.getLogger(__name__)


class SubmodelRepoImplementation(BaseModel):
    """Implementation of Submodel related API calls."""

    def __init__(self, client: "AasHttpClient"):
        """Initializes the SmImplementation with the given parameters."""
        self._client = client

        session = client.get_session()
        if session is None:
            raise ValueError(
                "HTTP session is not initialized in the client. Call 'initialize()' method of the client before creating SubmodelRegistryImplementation instance."
            )

        self._session: requests.Session = session

    # GET /submodels/{submodelIdentifier}
    def get_submodel_by_id(self, submodel_identifier: str, level: str = "", extent: str = "") -> dict | None:
        """Returns a specific Submodel.

        :param submodel_identifier: The Submodels unique id
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}"

        params: dict[str, str] = {}
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PUT /submodels/{submodelIdentifier}
    def put_submodels_by_id(self, submodel_identifier: str, request_body: dict) -> bool:
        """Updates a existing Submodel.

        :param submodel_identifier: The Submodels unique id
        :param request_body: Json data of the Submodel to update
        :return: True if the update was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.put(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # DELETE /submodels/{submodelIdentifier}
    def delete_submodel_by_id(self, submodel_identifier: str) -> bool:
        """Deletes a Submodel.

        :param submodel_identifier: The Submodels unique id
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def get_submodel_element_by_path_submodel_repo(
        self, submodel_identifier: str, id_short_path: str, level: str = "", extent: str = ""
    ) -> dict | None:
        """Returns a specific submodel element from the Submodel at a specified path.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel element data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}"

        params: dict[str, str] = {}
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PUT /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def put_submodel_element_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str, request_body: dict, level: str = "") -> bool:
        """Creates or updates an existing submodel element at a specified path within submodel elements hierarchy.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param request_body: Data for the submodel element
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: True if the operation was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}"

        params: dict[str, str] = {}
        if level:
            params["level"] = level

        self._client.set_token()

        try:
            response = self._session.put(url, json=request_body, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def post_submodel_element_by_path_submodel_repo(
        self, submodel_identifier: str, id_short_path: str, request_body: dict, level: str = "", extent: str = ""
    ) -> dict | None:
        """Creates a new submodel element at a specified path within submodel elements hierarchy.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param request_body: Data for the new Submodel element
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel element data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}"

        params: dict[str, str] = {}
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # DELETE /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def delete_submodel_element_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> bool:
        """Deletes a submodel element at a specified path within the submodel elements hierarchy.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}"

        self._client.set_token()
        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /submodels
    def get_all_submodels(
        self, semantic_id: str = "", id_short: str = "", limit: int = 0, cursor: str = "", level: str = "", extent: str = ""
    ) -> dict | None:
        """Returns all Submodels.

        :param semantic_id: The value of the semantic id reference (UTF8-BASE64-URL-encoded)
        :param id_short: The Submodels IdShort
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: List of Submodel data or None if an error occurred
        """
        url = f"{self._client.base_url}/submodels"

        params: dict[str, str] = {}
        if semantic_id:
            params["semanticId"] = semantic_id
        if id_short:
            params["idShort"] = id_short
        if limit:
            params["limit"] = str(limit)
        if cursor:
            params["cursor"] = cursor
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /submodels
    def post_submodel(self, request_body: dict) -> dict | None:
        """Creates a new Submodel.

        :param request_body: Json data of the Submodel to post
        :return: Submodel data or None if an error occurred
        """
        url = f"{self._client.base_url}/submodels"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # GET /submodels/{submodelIdentifier}/submodel-elements
    def get_all_submodel_elements_submodel_repository(
        self, submodel_identifier: str, limit: int = 100, cursor: str = "", level: str = "", extent: str = ""
    ) -> dict | None:
        """Returns all submodel elements including their hierarchy.

        :param submodel_identifier: The Submodels unique id
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: List of Submodel element data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements"

        params: dict[str, str] = {}
        if limit:
            params["limit"] = str(limit)
        if cursor:
            params["cursor"] = cursor
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /submodels/{submodelIdentifier}/submodel-elements
    def post_submodel_element_submodel_repo(self, submodel_identifier: str, request_body: dict) -> dict | None:
        """Creates a new submodel element.

        :param submodel_identifier: The Submodels unique id
        :param request_body: Data for the new Submodel element
        :return: Submodel element data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/invoke
    def invoke_operation_submodel_repo(self, submodel_identifier: str, id_short_path: str, request_body: dict, async_: str = "async") -> dict | None:
        """Synchronously invokes an Operation at a specified path.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the operation element (dot-separated)
        :param request_body: Input parameters for the operation
        :param async_: Determines whether an operation invocation is performed asynchronously or synchronously
        :return: Operation result or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/invoke"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # GET /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/$value
    def get_submodel_element_by_path_value_only_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> str | None:
        """Retrieves the value of a specific SubmodelElement.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: Submodel element value or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/$value"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PATCH /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/$value
    def patch_submodel_element_by_path_value_only_submodel_repo(
        self, submodel_identifier: str, id_short_path: str, value: str, level: str = ""
    ) -> bool:
        """Updates the value of an existing SubmodelElement.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param value: Submodel element value to update as string
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: True if the patch was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/$value"

        params: dict[str, str] = {}
        if level:
            params["level"] = level

        self._client.set_token()

        try:
            response = self._session.patch(url, json=value, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /submodels/{submodelIdentifier}/$value
    def get_submodel_by_id_value_only(self, submodel_identifier: str, level: str = "", extent: str = "") -> dict | None:
        """Returns a specific Submodel in the ValueOnly representation.

        :param submodel_identifier: The Submodels unique id
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel value as dict or None if an error occurred
        """
        params: dict[str, str] = {}
        if level:
            params["level"] = level
        if extent:
            params["extent"] = extent

        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/$value"

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PATCH /submodels/{submodelIdentifier}/$value
    def patch_submodel_by_id_value_only(self, submodel_identifier: str, request_body: dict, level: str = "") -> bool:
        """Updates the values of an existing Submodel.

        :param submodel_identifier: The Submodels unique id
        :param request_body: Submodel values to update as dict
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: True if the patch was successful, False otherwise
        """
        params: dict[str, str] = {}
        if level:
            params["level"] = level

        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/$value"

        self._client.set_token()

        try:
            response = self._session.patch(url, json=request_body, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /submodels/{submodelIdentifier}/$metadata
    def get_submodel_by_id_metadata(self, submodel_identifier: str, level: str = "") -> dict | None:
        """Returns the metadata attributes of a specific Submodel.

        :param submodel_identifier: The Submodels unique id
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: Metadata attributes of the Submodel as dict or None if an error occurred
        """
        params: dict[str, str] = {}
        if level:
            params["level"] = level

        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/$metadata"

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # not supported by Java Server

    # PATCH /submodels/{submodelIdentifier}
    def patch_submodel_by_id(self, submodel_identifier: str, submodel_data: dict) -> bool:
        """Updates an existing Submodel.

        :param submodel_identifier: The Submodels unique id
        :return: True if the patch was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.patch(url, json=submodel_data, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True



"""AAS HTTP Client Package."""

import importlib.metadata
from datetime import datetime, timezone

from aas_http_client.classes.client import aas_client
from aas_http_client.classes.client.aas_client import AasHttpClient
from aas_http_client.classes.wrapper import sdk_wrapper
from aas_http_client.classes.wrapper.sdk_wrapper import SdkWrapper
from aas_http_client.utilities import encoder, model_builder, sdk_tools
from aas_http_client.utilities.version_check import check_for_update

__copyright__ = f"Copyright (C) {datetime.now(tz=timezone.utc).year} :em engineering methods AG. All rights reserved."
__author__ = "Daniel Klein"

try:
    __license__ = "MIT"
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0-dev"

__project__ = "aas-http-client"
__package__ = "aas-http-client"

check_for_update()

__all__ = [
    "AasHttpClient",
    "SdkWrapper",
    "aas_client",
    "encoder",
    "model_builder",
    "sdk_tools",
    "sdk_wrapper",
]



"""Helper methods for HTTP operations."""

import json
import logging

from requests.models import Response

_logger = logging.getLogger(__name__)

STATUS_CODE_200 = 200
STATUS_CODE_201 = 201
STATUS_CODE_202 = 202
STATUS_CODE_204 = 204
STATUS_CODE_404 = 404


def log_response(response: Response, log_level: int = logging.ERROR):  # noqa: C901, PLR0912
    """Extracts and logs error messages from an HTTP response.

    This method parses the response content for error details, messages, or error fields,
    and logs each error message found. If the response cannot be decoded as JSON,
    it logs the raw response content. Always logs the HTTP status code.

    :param response: The HTTP response object to extract and log errors from
    :param log_level: The logging level to use (default is logging.ERROR)
    """
    result_error_messages: list[str] = []

    try:
        response_content_dict: dict = json.loads(response.content)

        if "detail" in response_content_dict:
            detail: dict = response_content_dict.get("detail", {})
            if "error" in detail:
                error: str = detail.get("error", "")
                result_error_messages.append(f"{error}")
            else:
                result_error_messages.append(f"{detail}")

        elif "messages" in response_content_dict or "Messages" in response_content_dict:
            messages: list = response_content_dict.get("messages", [])

            if not messages:
                messages = response_content_dict.get("Messages", [])

            for message in messages:
                if isinstance(message, dict) and "message" in message:
                    result_error_messages.append(message["message"])
                else:
                    result_error_messages.append(str(message))
        elif "error" in response_content_dict:
            result_error_messages.append(response_content_dict.get("error", ""))

        if len(result_error_messages) == 0 and response.text:
            result_error_messages.append(response.text)

    except json.JSONDecodeError:
        if response.content and response.content != "b''":
            result_error_messages.append(str(response.content))

    _logger.log(log_level, f"Status code: {response.status_code}")
    for result_error_message in result_error_messages:
        _logger.log(log_level, result_error_message)

    _logger.debug(f"Full response content: {str(response.content)}")



"""Attachment wrapper class for AAS HTTP Client."""

from dataclasses import dataclass


@dataclass(frozen=True)
class Attachment:
    """Represents an attachment with its content and metadata."""

    content: bytes
    content_type: str
    filename: str | None = None



import pytest
from pathlib import Path
from aas_http_client.classes.client.aas_client import create_by_config, AasHttpClient
from basyx.aas import model
from aas_http_client.utilities import encoder, sdk_tools, model_builder
import logging
from aas_http_client.demo.logging_handler import initialize_logging
import random

logger = logging.getLogger(__name__)

JAVA_SERVER_PORTS = [8075]
PYTHON_SERVER_PORTS = [5080, 80]

SM_ID = "fluid40/sm_http_client_unit_tests"
SHELL_ID = "fluid40/aas_http_client_unit_tests"

CONFIG_FILE_ENV = "./tests/server_configs/test_java_server_config.yml"
CONFIG_FILE_AAS_REG_ENV = "./tests/server_configs/test_aas_reg_server_config.yml"
CONFIG_FILE_SM_REG_ENV = "./tests/server_configs/test_sm_reg_server_config.yml"

shared_shell_descriptor: dict = {}
shared_sm_descriptor: dict = {}

@pytest.fixture(scope="module")
def client() -> AasHttpClient:
    try:
        initialize_logging()
        client = create_by_config(Path(CONFIG_FILE_ENV))

        rand = random.randint(0, 10)
        if (rand % 2) == 0:
            client.encoded_ids = True

    except Exception as e:
        raise RuntimeError("Unable to connect to server.")

    shells = client.shells.get_all_asset_administration_shells()
    if shells is None:
        raise RuntimeError("No shells found on server. Please check the server configuration.")

    return client

@pytest.fixture(scope="module")
def client_aas_reg(client: AasHttpClient) -> AasHttpClient:
    try:
        initialize_logging()
        reg_client = create_by_config(Path(CONFIG_FILE_AAS_REG_ENV))
        reg_client.encoded_ids = client.encoded_ids
    except Exception as e:
        raise RuntimeError("Unable to connect to server.")

    descriptors = reg_client.shell_registry.get_all_asset_administration_shell_descriptors()
    if descriptors is None:
        raise RuntimeError("No descriptors found on server. Please check the server configuration.")

    return reg_client

@pytest.fixture(scope="module")
def client_sm_reg(client: AasHttpClient) -> AasHttpClient:
    try:
        initialize_logging()
        reg_client = create_by_config(Path(CONFIG_FILE_SM_REG_ENV))
        reg_client.encoded_ids = client.encoded_ids
    except Exception as e:
        raise RuntimeError("Unable to connect to server.")

    descriptors = reg_client.submodel_registry.get_all_submodel_descriptors()
    if descriptors is None:
        raise RuntimeError("No descriptors found on server. Please check the server configuration.")

    return reg_client

@pytest.fixture(scope="module")
def shared_sm() -> model.Submodel:
    # create a Submodel
    return model_builder.create_base_submodel(identifier=SM_ID, id_short="sm_http_client_unit_tests", display_name="Submodel HTTP Client Unit Tests", description="This is a sample Submodel created for unit testing of the AAS HTTP Client.")

@pytest.fixture(scope="module")
def shared_aas(shared_sm: model.Submodel) -> model.AssetAdministrationShell:
    # create an AAS
    aas = model_builder.create_base_aas(identifier=SHELL_ID, id_short="aas_http_client_unit_tests", global_asset_identifier=SHELL_ID, display_name="AAS HTTP Client Unit Tests", description="This is a sample AAS created for unit testing of the AAS HTTP Client.")

    # add Submodel to AAS
    sdk_tools.add_submodel_to_aas(aas, shared_sm)

    return aas

@pytest.fixture(scope="module")
def global_shell_descriptor() -> dict:
    return shared_shell_descriptor

@pytest.fixture(scope="module")
def global_sm_descriptor() -> dict:
    return shared_sm_descriptor

def test_000a_clean_server(client: AasHttpClient, client_aas_reg: AasHttpClient, client_sm_reg: AasHttpClient):
    shells_result = client.shells.get_all_asset_administration_shells()

    for shell in shells_result.get("result", []):

        shell_id = shell["id"]

        if client.encoded_ids:
            shell_id = encoder.encode_base_64(shell_id)

        client.shells.delete_asset_administration_shell_by_id(shell_id)

    submodels_result = client.submodels.get_all_submodels()
    for submodel in submodels_result.get("result", []):

        sm_id = submodel["id"]

        if client.encoded_ids:
            sm_id = encoder.encode_base_64(sm_id)

        client.submodels.delete_submodel_by_id(sm_id)

    client_aas_reg.shell_registry.delete_all_asset_administration_shell_descriptors()
    client_sm_reg.submodel_registry.delete_all_submodel_descriptors()

    shells_result = client.shells.get_all_asset_administration_shells()
    assert len(shells_result.get("result")) == 0
    submodels_result = client.submodels.get_all_submodels()
    assert len(submodels_result.get("result")) == 0
    shell_descriptors_result = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert len(shell_descriptors_result.get("result")) == 0
    sm_descriptors_result = client_sm_reg.submodel_registry.get_all_submodel_descriptors()
    assert len(sm_descriptors_result.get("result")) == 0

def test_000b_post_assets(client: AasHttpClient, shared_aas: model.AssetAdministrationShell, shared_sm: model.Submodel):
    sm_data = sdk_tools.convert_to_dict(shared_sm)
    sm_result = client.submodels.post_submodel(sm_data)

    assert sm_result is not None

    shell_data = sdk_tools.convert_to_dict(shared_aas)
    shell_result = client.shells.post_asset_administration_shell(shell_data)
    assert shell_result is not None

def test_001a_get_self_description_shell(client_aas_reg: AasHttpClient):
    description = client_aas_reg.shell_registry.get_self_description()

    assert description is not None
    assert "profiles" in description
    assert len(description["profiles"]) == 2

def test_001b_get_self_description_sm(client_sm_reg: AasHttpClient):
    description = client_sm_reg.submodel_registry.get_self_description()

    assert description is not None
    assert "profiles" in description
    assert len(description["profiles"]) == 2

def test_002_search(client_aas_reg: AasHttpClient):
    request_body = {
        "page": {
            "index": 0,
            "size": 1
        },
        "sortBy": {
            "direction": "ASC",
            "path": [
            "idShort"
            ]
        }
    }

    search_result = client_aas_reg.shell_registry.search(request_body)

    assert search_result is not None
    assert "total" in search_result
    total = search_result["total"]
    assert total == 1
    assert "hits" in search_result
    hits = search_result["hits"]
    assert hits is not None
    assert len(hits) == 1
    assert hits[0]["id"] == SHELL_ID

def test_003_get_all_asset_administration_shell_descriptors(client_aas_reg: AasHttpClient):
    descriptors = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()

    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1
    assert results[0]["id"] == SHELL_ID

    global shared_shell_descriptor
    shared_shell_descriptor.clear()
    shared_shell_descriptor.update(results[0])

def test_004_get_all_submodel_descriptors(client_sm_reg: AasHttpClient):
    descriptors = client_sm_reg.submodel_registry.get_all_submodel_descriptors()

    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1
    assert results[0]["id"] == SM_ID

    global shared_sm_descriptor
    shared_sm_descriptor.clear()
    shared_sm_descriptor.update(results[0])

def test_005a_get_endpoint(client: AasHttpClient, client_sm_reg: AasHttpClient):
    sm_id = SM_ID
    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    descriptor = client_sm_reg.submodel_registry.get_submodel_descriptor_by_id(sm_id)
    assert descriptor is not None
    assert descriptor.get("id") == SM_ID
    endpoints = descriptor.get("endpoints", [])
    assert len(endpoints) == 1
    endpoint = endpoints[0]
    assert "protocolInformation" in endpoint
    protocol_info: dict = endpoint["protocolInformation"]
    assert protocol_info.get("endpointProtocol") == "http"
    href = protocol_info.get("href", "")
    assert href is not None

    print(f"Testing endpoint href: {href}")

    sm = client.get_endpoint(href)
    assert sm is not None
    assert sm.get("id") == SM_ID

def test_005b_delete_assets(client: AasHttpClient, client_aas_reg: AasHttpClient, client_sm_reg: AasHttpClient):
    sm_id = SM_ID
    shell_id = SHELL_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.submodels.delete_submodel_by_id(sm_id)
    assert result

    submodels = client.shells.delete_asset_administration_shell_by_id(shell_id)
    assert submodels

    shells_result = client.shells.get_all_asset_administration_shells()
    assert len(shells_result.get("result")) == 0
    submodels_result = client.submodels.get_all_submodels()
    assert len(submodels_result.get("result")) == 0
    shell_descriptors_result = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert len(shell_descriptors_result.get("result")) == 0
    sm_descriptors_result = client_sm_reg.submodel_registry.get_all_submodel_descriptors()
    assert len(sm_descriptors_result.get("result")) == 0

def test_006_post_asset_administration_shell_descriptor(client_aas_reg: AasHttpClient, global_shell_descriptor: dict):
    result = client_aas_reg.shell_registry.post_asset_administration_shell_descriptor(global_shell_descriptor)

    assert result is not None
    assert "id" in result
    assert result["id"] == SHELL_ID

    descriptors = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1
    assert results[0]["id"] == SHELL_ID

def test_007_get_asset_administration_shell_descriptor_by_id(client_aas_reg: AasHttpClient):
    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    descriptor = client_aas_reg.shell_registry.get_asset_administration_shell_descriptor_by_id(shell_id)

    assert descriptor is not None
    assert descriptor["id"] == SHELL_ID

def test_008_put_asset_administration_shell_descriptor_by_id(client_aas_reg: AasHttpClient, global_shell_descriptor: dict):
    global_shell_descriptor["idShort"] = "sm_http_client_unit_tests_updated"

    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client_aas_reg.shell_registry.put_asset_administration_shell_descriptor_by_id(shell_id, global_shell_descriptor)

    assert result is True

    descriptor = client_aas_reg.shell_registry.get_asset_administration_shell_descriptor_by_id(shell_id)
    assert descriptor is not None
    assert descriptor["id"] == SHELL_ID
    assert descriptor["idShort"] ==  global_shell_descriptor["idShort"]

def test_009_delete_asset_administration_shell_descriptor_by_id(client_aas_reg: AasHttpClient, global_shell_descriptor: dict):
    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client_aas_reg.shell_registry.delete_asset_administration_shell_descriptor_by_id(shell_id)

    assert result is True

    descriptors = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 0

    # Re-post the descriptor for further tests
    global_shell_descriptor["idShort"] = "sm_http_client_unit_tests"
    result = client_aas_reg.shell_registry.post_asset_administration_shell_descriptor(global_shell_descriptor)

    descriptors = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1

def test_010_post_submodel_descriptor(client_sm_reg: AasHttpClient, global_sm_descriptor: dict):
    result = client_sm_reg.submodel_registry.post_submodel_descriptor(global_sm_descriptor)

    assert result is not None
    assert "id" in result
    assert result["id"] == SM_ID

    descriptors = client_sm_reg.submodel_registry.get_all_submodel_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1
    assert results[0]["id"] == SM_ID

def test_011a_delete_all_asset_administration_shell_descriptors(client_aas_reg: AasHttpClient):
    result = client_aas_reg.shell_registry.delete_all_asset_administration_shell_descriptors()
    assert result

    descriptors = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 0

def test_011b_delete_all_submodel_descriptors(client_sm_reg: AasHttpClient):
    result = client_sm_reg.submodel_registry.delete_all_submodel_descriptors()
    assert result

    descriptors = client_sm_reg.submodel_registry.get_all_submodel_descriptors()
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 0

def test_011c_post_aas_descriptor(client_aas_reg: AasHttpClient, global_shell_descriptor: dict, global_sm_descriptor: dict):
    global_shell_descriptor["submodelDescriptors"] = []
    sm_descriptors: list[dict] = global_shell_descriptor.get("submodelDescriptors", [])
    sm_descriptors.append(global_sm_descriptor)

    result = client_aas_reg.shell_registry.post_asset_administration_shell_descriptor(global_shell_descriptor)
    assert result is not None
    assert "id" in result
    assert result["id"] == SHELL_ID
    assert "submodelDescriptors" in result
    assert len(result["submodelDescriptors"]) == 1

    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    get_result = client_aas_reg.shell_registry.get_asset_administration_shell_descriptor_by_id(shell_id)
    assert get_result is not None
    assert "id" in get_result
    assert get_result["id"] == SHELL_ID
    assert "submodelDescriptors" in get_result
    assert len(get_result["submodelDescriptors"])

def test_012_get_submodel_descriptor_by_id_through_superpath(client_aas_reg: AasHttpClient):
    shell_id = SHELL_ID
    sm_id = SM_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    descriptor = client_aas_reg.shell_registry.get_submodel_descriptor_by_id_through_superpath(shell_id, sm_id)
    assert descriptor is not None
    assert descriptor["id"] == SM_ID

def test_013_put_submodel_descriptor_by_id_through_superpath(client_aas_reg: AasHttpClient, global_sm_descriptor: dict):
    global_sm_descriptor["idShort"] = "sm_http_client_unit_tests_updated"

    shell_id = SHELL_ID
    sm_id = SM_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = client_aas_reg.shell_registry.put_submodel_descriptor_by_id_through_superpath(shell_id, sm_id, global_sm_descriptor)
    assert result is True

    descriptor = client_aas_reg.shell_registry.get_submodel_descriptor_by_id_through_superpath(shell_id, sm_id)
    assert descriptor is not None
    assert descriptor["id"] == SM_ID
    assert descriptor["idShort"] == "sm_http_client_unit_tests_updated"

def test_014_delete_submodel_descriptor_by_id_through_superpath(client_aas_reg: AasHttpClient):
    shell_id = SHELL_ID
    sm_id = SM_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = client_aas_reg.shell_registry.delete_submodel_descriptor_by_id_through_superpath(shell_id, sm_id)
    assert result is True

    sm_descriptors = client_aas_reg.shell_registry.get_submodel_descriptor_by_id_through_superpath(shell_id, sm_id)
    assert sm_descriptors is None

    shell_descriptor = client_aas_reg.shell_registry.get_asset_administration_shell_descriptor_by_id(shell_id)
    assert shell_descriptor is not None
    assert "submodelDescriptors" in shell_descriptor
    assert len(shell_descriptor["submodelDescriptors"]) == 0

def test_015_post_submodel_descriptor_through_superpath(client_aas_reg: AasHttpClient, global_sm_descriptor: dict):
    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    descriptor = client_aas_reg.shell_registry.post_submodel_descriptor_through_superpath(shell_id, global_sm_descriptor)
    assert descriptor is not None
    assert "id" in descriptor
    assert descriptor["id"] == SM_ID

def test_016_get_all_submodel_descriptors_through_superpath(client_aas_reg: AasHttpClient):
    shell_id = SHELL_ID

    if client_aas_reg.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    descriptors = client_aas_reg.shell_registry.get_all_submodel_descriptors_through_superpath(shell_id)
    assert descriptors is not None
    assert "result" in descriptors
    results = descriptors["result"]
    assert results is not None
    assert len(results) == 1
    assert results[0]["id"] == SM_ID

def test_017_post_submodel_descriptor(client_sm_reg: AasHttpClient, global_sm_descriptor: dict):
    result = client_sm_reg.submodel_registry.post_submodel_descriptor(global_sm_descriptor)
    assert result is not None
    assert "id" in result
    assert result["id"] == SM_ID

def test_018_get_submodel_descriptor_by_id(client_sm_reg: AasHttpClient):
    sm_id = SM_ID

    if client_sm_reg.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    descriptor = client_sm_reg.submodel_registry.get_submodel_descriptor_by_id(sm_id)
    assert descriptor is not None
    assert descriptor["id"] == SM_ID

def test_019_put_submodel_descriptor_by_id(client_sm_reg: AasHttpClient, global_sm_descriptor: dict):
    global_sm_descriptor["idShort"] = "sm_http_client_unit_tests_updated"

    sm_id = SM_ID

    if client_sm_reg.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client_sm_reg.submodel_registry.put_submodel_descriptor_by_id(sm_id, global_sm_descriptor)
    assert result is True

    descriptor = client_sm_reg.submodel_registry.get_submodel_descriptor_by_id(sm_id)
    assert descriptor is not None
    assert descriptor["id"] == SM_ID
    assert descriptor["idShort"] == "sm_http_client_unit_tests_updated"

def test_020_delete_submodel_descriptor_by_id(client_sm_reg: AasHttpClient):
    sm_id = SM_ID

    if client_sm_reg.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client_sm_reg.submodel_registry.delete_submodel_descriptor_by_id(sm_id)
    assert result is True

    sm_descriptors = client_sm_reg.submodel_registry.get_submodel_descriptor_by_id(sm_id)
    assert sm_descriptors is None

def test_099_cleanup(client: AasHttpClient, client_aas_reg: AasHttpClient, client_sm_reg: AasHttpClient):
    shells_result = client.shells.get_all_asset_administration_shells()

    for shell in shells_result.get("result", []):
        client.shells.delete_asset_administration_shell_by_id(shell["id"])

    submodels_result = client.submodels.get_all_submodels()
    for submodel in submodels_result.get("result", []):
        client.submodels.delete_submodel_by_id(submodel["id"])

    client_aas_reg.shell_registry.delete_all_asset_administration_shell_descriptors()
    client_sm_reg.submodel_registry.delete_all_submodel_descriptors()

    shells_result = client.shells.get_all_asset_administration_shells()
    assert len(shells_result.get("result")) == 0
    submodels_result = client.submodels.get_all_submodels()
    assert len(submodels_result.get("result")) == 0
    shell_descriptors_result = client_aas_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    assert len(shell_descriptors_result.get("result")) == 0
    sm_descriptors_result = client_sm_reg.submodel_registry.get_all_submodel_descriptors()
    assert len(sm_descriptors_result.get("result")) == 0



import pytest
from aas_http_client.utilities import encoder

def test_001a_decode_base_64():
    encoded = "RW5lcmd5TW9uaXRvcmluZw"

    decoded = encoder.decode_base_64(encoded)
    assert decoded == "EnergyMonitoring"
    new_encoded = encoder.encode_base_64(decoded)
    assert new_encoded == encoded


def test_001b_decode_base_64():
    encoded = "aHR0cHM6Ly9mbHVpZDQwLmRlL2lkcy9zbS8xNTE1XzUzMDZfNzYyNF81MzY3"

    decoded = encoder.decode_base_64(encoded)
    assert decoded == "https://fluid40.de/ids/sm/1515_5306_7624_5367"
    new_encoded = encoder.encode_base_64(decoded)
    assert new_encoded == encoded

def test_002a_encode_base_64():
    decoded = "EnergyMonitoring"

    encoded = encoder.encode_base_64(decoded)
    assert encoded == "RW5lcmd5TW9uaXRvcmluZw"
    new_decoded = encoder.decode_base_64(encoded)
    assert new_decoded == decoded

def test_002b_encode_base_64():
    decoded = "https://fluid40.de/ids/sm/1515_5306_7624_5367"

    encoded = encoder.encode_base_64(decoded)
    assert encoded == "aHR0cHM6Ly9mbHVpZDQwLmRlL2lkcy9zbS8xNTE1XzUzMDZfNzYyNF81MzY3"
    new_decoded = encoder.decode_base_64(encoded)
    assert new_decoded == decoded



"""Basic Authentication Configuration."""

from pydantic import BaseModel, Field, PrivateAttr


class BearerAuth(BaseModel):
    """Bearer Authentication Configuration.

    :param BaseModel: Pydantic BaseModel for data validation.
    """

    _token: str = PrivateAttr(default="")

    def set_token(self, token: str) -> None:
        """Set the bearer token for the authentication.

        :param token: Bearer token for the authentication.
        """
        self._token = token

    def get_token(self) -> str:
        """Get the bearer token for the authentication.

        :return: The bearer token.
        """
        return self._token

    def is_active(self) -> bool:
        """Check if the bearer authentication is active.

        :return: True if the token is not empty, False otherwise.
        """
        return bool(self._token)


class BasicAuth(BaseModel):
    """Basic Authentication Configuration.

    :param BaseModel: Pydantic BaseModel for data validation.
    """

    username: str = Field(default="", alias="Username", description="Username for the basic authentication.")
    _password: str = PrivateAttr(default="")

    def is_active(self) -> bool:
        """Check if the basic authentication is active.

        :return: True if the username is not empty, False otherwise.
        """
        return bool(self.username and self._password)

    def set_password(self, password: str) -> None:
        """Set the password for the basic authentication.

        :param password: Password for the basic authentication.
        """
        self._password = password

    def get_password(self) -> str:
        """Get the password for the basic authentication.

        :return: The password.
        """
        return self._password


class OAuth(BaseModel):
    """Open Authentication Configuration.

    :param BaseModel: Pydantic BaseModel for data validation.
    """

    token_url: str = Field(default="", alias="TokenUrl", description="Endpoint URL for the token request.")
    client_id: str = Field(default="", alias="ClientId", description="Client identifier for authentication.")
    grant_type: str = Field(default="client_credentials", alias="GrantType", description="Grant type for the authentication.")
    header_name: str = Field(default="Authorization", alias="HeaderName", description="Header name for the authentication.")
    _client_secret: str = PrivateAttr(default="")

    def is_active(self) -> bool:
        """Check if the service provider authentication is active.

        :return: True if the client ID is not empty, False otherwise.
        """
        return bool(self.client_id and self._client_secret and self.token_url)

    def set_client_secret(self, client_secret: str) -> None:
        """Set the client secret for the authentication.

        :param client_secret: Client secret for the authentication.
        """
        self._client_secret = client_secret

    def get_client_secret(self) -> str:
        """Get the client secret for the authentication.

        :return: The client secret.
        """
        if self._client_secret is None:
            return ""

        return self._client_secret


class AuthenticationConfig(BaseModel):
    """Authentication Configuration.

    param BaseModel: Pydantic BaseModel for data validation.
    """

    basic_auth: BasicAuth = Field(default_factory=BasicAuth, alias="BasicAuth", description="Basic authentication configuration.")
    o_auth: OAuth = Field(
        default_factory=OAuth,
        alias="OAuth",
        description="Service provider authentication configuration.",
    )
    bearer_auth: BearerAuth = Field(default_factory=BearerAuth, alias="BearerAuth", description="Bearer authentication configuration.")



"""Implements authentication methods for the HTTP client."""

import json
import logging
import time
from enum import Enum

import requests
from requests.auth import HTTPBasicAuth

from aas_http_client.classes.Configuration.config_classes import OAuth
from aas_http_client.utilities.http_helper import log_response

_logger = logging.getLogger(__name__)


class AuthMethod(Enum):
    """Defines authentication methods.

    :param Enum: Base class for enumerations
    """

    No = 1
    basic_auth = 2
    o_auth = 3
    bearer = 4


class TokenData:
    """Holds token data."""

    def __init__(self, access_token: str, token_type: str, token_expiry: float):
        """Initializes the TokenData with the given parameters."""
        self.access_token: str = access_token
        self.token_type: str = token_type
        self.token_expiry: float = token_expiry


def get_token(o_auth_configuration: OAuth) -> TokenData | None:
    """Get token based on the provided OAuth configuration.

    :param auth_configuration: Authentication configuration
    :return: Access token or None if an error occurred
    """
    if o_auth_configuration.grant_type == "password":
        token = get_token_by_password(
            o_auth_configuration.token_url,
            o_auth_configuration.client_id,
            o_auth_configuration.get_client_secret(),
        )

    elif o_auth_configuration.is_active() and o_auth_configuration.grant_type == "client_credentials":
        token = get_token_by_basic_auth(
            o_auth_configuration.token_url,
            o_auth_configuration.client_id,
            o_auth_configuration.get_client_secret(),
        )

    if not token:
        _logger.error(f"Failed to receive token from endpoint '{o_auth_configuration.token_url}'")
        return None

    return token


def get_token_by_basic_auth(endpoint: str, username: str, password: str, timeout=200) -> TokenData | None:
    """Get token from a specific authentication service provider by basic authentication.

    :param endpoint: Get token endpoint for the authentication service provider
    :param username: Username for the authentication service provider
    :param password: Password for the authentication service provider
    :param timeout: Timeout for the API calls, defaults to 200
    :return: Access token or None if an error occurred
    """
    data = {"grant_type": "client_credentials"}

    auth = HTTPBasicAuth(username, password)

    return __get_token_from_endpoint(endpoint, data, auth, timeout)


def get_token_by_password(endpoint: str, username: str, password: str, timeout=200) -> TokenData | None:
    """Get token from a specific authentication service provider by username and password.

    :param endpoint: Get token endpoint for the authentication service provider
    :param username: Username for the authentication service provider
    :param password: Password for the authentication service provider
    :param timeout: Timeout for the API calls, defaults to 200
    :return: Access token or None if an error occurred
    """
    data = {"grant_type": "password", "username": username, "password": password}

    return __get_token_from_endpoint(endpoint, data, None, timeout)


def __get_token_from_endpoint(endpoint: str, data: dict[str, str], auth: HTTPBasicAuth | None = None, timeout: int = 200) -> TokenData | None:
    """Get token from a specific authentication service provider.

    :param endpoint: Get token endpoint for the authentication service provider
    :param data: Data for the authentication service provider
    :param timeout: Timeout for the API calls, defaults to 200
    :return: Access token or None if an error occurred
    """
    try:
        response = requests.post(endpoint, auth=auth, data=data, timeout=timeout)
        _logger.debug(f"Call REST API url '{response.url}'")

        if response.status_code != 200:
            log_response(response)
            return None

    except requests.exceptions.RequestException as e:
        _logger.error(f"Error call REST API: {e}")
        return None

    content = response.content.decode("utf-8")

    if not content:
        _logger.error("No content in token response")
        return None

    data = json.loads(content)

    if not data:
        _logger.error("No data in token response")
        return None

    access_token: str = data.get("access_token", "").strip()
    expires_in: int = int(data.get("expires_in", "0"))
    if not access_token or not expires_in:
        _logger.error("Invalid token data in response")
        return None

    token_type: str = data.get("token_type", "").strip()
    now: float = time.time()
    token_expiry: float = now + int(expires_in) - 60  # Subtract 60 seconds as buffer

    return TokenData(access_token=access_token, token_type=token_type, token_expiry=token_expiry)



"""Pagination wrapper classes for AAS HTTP Client."""

import logging

from basyx.aas import model

from aas_http_client.utilities.sdk_tools import convert_to_object

_logger = logging.getLogger(__name__)


class PagingMetadata:
    """Class representing pagination metadata."""

    cursor: str

    def __init__(self, cursor: str) -> None:
        """Initialize a paging metadata object.

        :param cursor: Cursor for the next page
        """
        self.cursor = cursor


class ShellPaginatedData:
    """Class representing paginated data for Asset Administration Shells."""

    paging_metadata: PagingMetadata
    results: list[model.AssetAdministrationShell]

    def __init__(self, cursor: str, results: list[model.AssetAdministrationShell]) -> None:
        """Initialize a paginated data object.

        :param paging_metadata: Paging metadata
        :param results: list of results
        """
        self.paging_metadata = PagingMetadata(cursor)
        self.results = results


class ReferencePaginatedData:
    """Class representing paginated data for References."""

    paging_metadata: PagingMetadata
    results: list[model.ModelReference]

    def __init__(self, cursor: str, results: list[model.ModelReference]) -> None:
        """Initialize a paginated data object.

        :param paging_metadata: Paging metadata
        :param results: list of results
        """
        self.paging_metadata = PagingMetadata(cursor)
        self.results = results


class SubmodelPaginatedData:
    """Class representing paginated data for Submodels."""

    paging_metadata: PagingMetadata
    results: list[model.Submodel]

    def __init__(self, cursor: str, results: list[model.Submodel]) -> None:
        """Initialize a paginated data object.

        :param paging_metadata: Paging metadata
        :param results: list of results
        """
        self.paging_metadata = PagingMetadata(cursor)
        self.results = results


class SubmodelElementPaginatedData:
    """Class representing paginated data for Submodel Elements."""

    paging_metadata: PagingMetadata
    results: list[model.SubmodelElement]

    def __init__(self, cursor: str, results: list[model.SubmodelElement]) -> None:
        """Initialize a paginated data object.

        :param paging_metadata: Paging metadata
        :param results: list of results
        """
        self.paging_metadata = PagingMetadata(cursor)
        self.results = results


class ShellDescriptorPaginatedData:
    """Class representing paginated data for Shell Descriptors."""

    paging_metadata: PagingMetadata
    results: list[model.SubmodelElement]

    def __init__(self, cursor: str, results: list[model.SubmodelElement]) -> None:
        """Initialize a paginated data object.

        :param paging_metadata: Paging metadata
        :param results: list of results
        """
        self.paging_metadata = PagingMetadata(cursor)
        self.results = results


def create_shell_paging_data(content: dict) -> ShellPaginatedData | None:
    """Create a ShellPaginatedData object from a dictionary.

    :param content: Dictionary containing paginated shell data
    :return: ShellPaginatedData object
    """
    aas_list: list[model.AssetAdministrationShell] = []

    results: list = content.get("result", [])
    if not results or len(results) == 0:
        _logger.warning("No shells found on server.")
        return ShellPaginatedData(cursor="", results=[])

    for result in results:
        if not isinstance(result, dict):
            _logger.error(f"Invalid shell data: {result}")
            return None

        aas = convert_to_object(result)

        if aas:
            aas_list.append(aas)

    cursor = ""
    paging_metadata_dict = content.get("paging_metadata", {})

    if "cursor" in paging_metadata_dict:
        cursor = paging_metadata_dict["cursor"]

    return ShellPaginatedData(
        cursor=cursor,
        results=aas_list,
    )


def create_submodel_paging_data(content: dict) -> SubmodelPaginatedData | None:
    """Create a SubmodelPaginatedData object from a dictionary.

    :param content: Dictionary containing paginated submodel data
    :return: SubmodelPaginatedData object
    """
    sm_list: list[model.Submodel] = []

    results: list = content.get("result", [])
    if not results or len(results) == 0:
        _logger.warning("No shells found on server.")
        return SubmodelPaginatedData(cursor="", results=[])

    for result in results:
        if not isinstance(result, dict):
            _logger.error(f"Invalid shell data: {result}")
            return None

        sm = convert_to_object(result)

        if sm:
            sm_list.append(sm)

    cursor = ""
    paging_metadata_dict = content.get("paging_metadata", {})

    if "cursor" in paging_metadata_dict:
        cursor = paging_metadata_dict["cursor"]

    return SubmodelPaginatedData(
        cursor=cursor,
        results=sm_list,
    )


def create_submodel_element_paging_data(content: dict) -> SubmodelElementPaginatedData | None:
    """Create a SubmodelElementPaginatedData object from a dictionary.

    :param content: Dictionary containing paginated submodel element data
    :return: SubmodelElementPaginatedData object
    """
    sme_list: list[model.SubmodelElement] = []

    results: list = content.get("result", [])
    if not results or len(results) == 0:
        _logger.warning("No shells found on server.")
        return SubmodelElementPaginatedData(cursor="", results=[])

    for result in results:
        if not isinstance(result, dict):
            _logger.error(f"Invalid shell data: {result}")
            return None

        sme = convert_to_object(result)

        if sme:
            sme_list.append(sme)

    cursor = ""
    paging_metadata_dict = content.get("paging_metadata", {})

    if "cursor" in paging_metadata_dict:
        cursor = paging_metadata_dict["cursor"]

    return SubmodelElementPaginatedData(
        cursor=cursor,
        results=sme_list,
    )


def create_shell_descriptor_paging_data(content: dict) -> SubmodelElementPaginatedData | None:
    """Create a SubmodelElementPaginatedData object from a dictionary.

    :param content: Dictionary containing paginated submodel element data
    :return: SubmodelElementPaginatedData object
    """
    sme_list: list[model.SubmodelElement] = []

    results: list = content.get("result", [])
    if not results or len(results) == 0:
        _logger.warning("No shells found on server.")
        return SubmodelElementPaginatedData(cursor="", results=[])

    for result in results:
        if not isinstance(result, dict):
            _logger.error(f"Invalid shell data: {result}")
            return None

        sme = convert_to_object(result)

        if sme:
            sme_list.append(sme)

    cursor = ""
    paging_metadata_dict = content.get("paging_metadata", {})

    if "cursor" in paging_metadata_dict:
        cursor = paging_metadata_dict["cursor"]

    return SubmodelElementPaginatedData(
        cursor=cursor,
        results=sme_list,
    )


def create_reference_paging_data(content: dict) -> ReferencePaginatedData | None:
    """Create a ReferencePaginatedData object from a dictionary.

    :param content: Dictionary containing paginated reference data
    :return: ReferencePaginatedData object
    """
    ref_list: list[model.ModelReference] = []

    results: list = content.get("result", [])
    if not results or len(results) == 0:
        _logger.warning("No shells found on server.")
        return ReferencePaginatedData(cursor="", results=[])

    for result in results:
        if not isinstance(result, dict):
            _logger.error(f"Invalid shell data: {result}")
            return None

        reference = convert_to_object(result)

        if reference:
            ref_list.append(reference)

    cursor = ""
    paging_metadata_dict = content.get("paging_metadata", {})

    if "cursor" in paging_metadata_dict:
        cursor = paging_metadata_dict["cursor"]

    return ReferencePaginatedData(
        cursor=cursor,
        results=ref_list,
    )



"""Submodel Registry Implementation Module."""

import json
import logging
from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel

if TYPE_CHECKING:
    from aas_http_client.classes.client.aas_client import AasHttpClient


from aas_http_client.utilities.encoder import encode_base_64
from aas_http_client.utilities.http_helper import (
    STATUS_CODE_200,
    STATUS_CODE_201,
    STATUS_CODE_204,
    STATUS_CODE_404,
    log_response,
)

_logger = logging.getLogger(__name__)


class SubmodelRegistryImplementation(BaseModel):
    """Implementation of Submodel Registry related API calls."""

    def __init__(self, client: "AasHttpClient"):
        """Initializes the SubmodelRegistryImplementation with the given client."""
        self._client = client

        session = client.get_session()
        if session is None:
            raise ValueError(
                "HTTP session is not initialized in the client. Call 'initialize()' method of the client before creating SubmodelRegistryImplementation instance."
            )

        self._session: requests.Session = session

    # GET /submodel-descriptors/{submodelIdentifier}
    def get_submodel_descriptor_by_id(self, submodel_identifier: str) -> dict | None:
        """Returns the Submodel Descriptor for the given submodel identifier.

        :param submodel_identifier: The unique identifier of the Submodel Descriptor
        :return: Submodel Descriptor data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodel-descriptors/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel Descriptor with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PUT /submodel-descriptors/{submodelIdentifier}
    def put_submodel_descriptor_by_id(self, submodel_identifier: str, request_body: dict) -> bool:
        """Creates or updates an existing Submodel Descriptor.

        :param submodel_identifier: The unique identifier of the Submodel Descriptor
        :param request_body: Submodel Descriptor object
        :return: Updated Submodel Descriptor data or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodel-descriptors/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.put(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel Descriptor with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # DELETE /submodel-descriptors/{submodelIdentifier}
    def delete_submodel_descriptor_by_id(self, submodel_identifier: str) -> bool:
        """Deletes a Submodel Descriptor, i.e. de-registers a submodel.

        :param submodel_identifier: The unique identifier of the Submodel Descriptor
        :return: True if deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodel-descriptors/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel Descriptor with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /submodel-descriptors
    def get_all_submodel_descriptors(self, limit: int = 100, cursor: str = "") -> dict | None:
        """Returns all Submodel Descriptors.

        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :return: Submodel Descriptors data or None if an error occurred
        """
        url = f"{self._client.base_url}/submodel-descriptors"

        params: dict[str, str] = {}
        if limit:
            params["limit"] = str(limit)
        if cursor:
            params["cursor"] = cursor

        self._client.set_token()

        try:
            response = self._session.get(url, params=params, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /submodel-descriptors
    def post_submodel_descriptor(self, request_body: dict) -> dict | None:
        """Creates a new Submodel Descriptor, i.e. registers a submodel.

        :param request_body: Submodel Descriptor object
        :return: Created Submodel Descriptor data or None if an error occurred
        """
        url = f"{self._client.base_url}/submodel-descriptors"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # DELETE /submodel-descriptors
    def delete_all_submodel_descriptors(self) -> bool:
        """Deletes all Submodel Descriptors.

        :return: True if deletion was successful, False otherwise
        """
        url = f"{self._client.base_url}/submodel-descriptors"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /description
    def get_self_description(self) -> dict | None:
        """Returns the self-describing information of a network resource (ServiceDescription).

        :return: self-describing information of a network resource
        """
        url = f"{self._client.base_url}/description"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)



from multiprocessing.sharedctypes import Value
import pytest
from pathlib import Path
from aas_http_client.classes.client.aas_client import create_by_config, AasHttpClient, create_by_dict, create_by_url
from basyx.aas import model
import aas_http_client.utilities.model_builder as model_builder
import aas_http_client.utilities.sdk_tools as sdk_tools
import json
import basyx.aas.adapter.json
from urllib.parse import urlparse
import logging
from aas_http_client.demo.logging_handler import initialize_logging
from aas_http_client.utilities import encoder
import random
import json

logger = logging.getLogger(__name__)

JAVA_SERVER_PORTS = [8075]
PYTHON_SERVER_PORTS = [5080, 80]
DOTNET_SERVER_PORTS = [5043]

AIMC_SM_ID = "https://fluid40.de/ids/sm/7644_4034_2556_2369"
SM_ID = "fluid40/sm_http_client_unit_tests"
SHELL_ID = "fluid40/aas_http_client_unit_tests"

CONFIG_FILES = [
    "./tests/server_configs/test_java_server_config.yml",
    "./tests/server_configs/test_dotnet_server_config.yml",
    "./tests/server_configs/test_python_server_config.yml"
]

# CONFIG_FILES = [
#     "./tests/server_configs/test_dotnet_server_config_local.yml",
# ]

@pytest.fixture(params=CONFIG_FILES, scope="module")
def client(request) -> AasHttpClient:
    try:
        initialize_logging()
        file = Path(request.param).resolve()

        if not file.exists():
            raise FileNotFoundError(f"Configuration file {file} does not exist.")

        client = create_by_config(file)

        if client is None:
            raise RuntimeError("Failed to create client from configuration file.")

        # Randomly set encoded_ids to True or False for testing both scenarios
        rand = random.randint(0, 10)
        if (rand % 2) == 0:
            client.encoded_ids = True

    except Exception as e:
        raise RuntimeError("Unable to connect to server.")


    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shells = client.shells.get_all_asset_administration_shells()
    if shells is None:
        raise RuntimeError("No shells found on server. Please check the server configuration.")

    return client

@pytest.fixture(scope="module")
def shared_sme_string() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_string", model.datatypes.String, "Sample String Value", description="This is a sample string property for unit testing.", display_name="Sample String Property")

@pytest.fixture(scope="module")
def shared_sme_bool() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_bool", model.datatypes.Boolean, True, description="This is a sample boolean property for unit testing.", display_name="Sample Boolean Property")

@pytest.fixture(scope="module")
def shared_sme_int() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_int", model.datatypes.Integer, 262, description="This is a sample integer property for unit testing.", display_name="Sample Integer Property")

@pytest.fixture(scope="module")
def shared_sme_float() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_float", model.datatypes.Float, 262.3, description="This is a sample float property for unit testing.", display_name="Sample Float Property")

@pytest.fixture(scope="module")
def shared_sme_collection() -> model.SubmodelElementCollection:
    # create a Submodel
    values: list[model.SubmodelElement] = []
    values.append(model_builder.create_base_submodel_element_property("coll_element_1", model.datatypes.Integer, 262, description="This is a sample integer property for unit testing.", display_name="Sample Integer Property"))
    values.append(model_builder.create_base_submodel_element_property("coll_element_2", model.datatypes.String, "262", description="This is a sample string property for unit testing.", display_name="Sample String Property"))
    values.append(model_builder.create_base_submodel_element_property("coll_element_3", model.datatypes.Float, 262.3, description="This is a sample float property for unit testing.", display_name="Sample Float Property"))

    return model_builder.create_base_submodel_element_collection("sme_property_collection", values, description="This is a sample collection property for unit testing.", display_name="Sample Collection Property")

@pytest.fixture(scope="module")
def shared_sm() -> model.Submodel:
    # create a Submodel
    return model_builder.create_base_submodel(identifier=SM_ID, id_short="sm_http_client_unit_tests", display_name="Submodel HTTP Client Unit Tests", description="This is a sample Submodel created for unit testing of the AAS HTTP Client.")

@pytest.fixture(scope="module")
def shared_aas(shared_sm: model.Submodel) -> model.AssetAdministrationShell:
    # create an AAS
    aas = model_builder.create_base_aas(identifier=SHELL_ID, id_short="aas_http_client_unit_tests", global_asset_identifier=SHELL_ID, display_name="AAS HTTP Client Unit Tests", description="This is a sample AAS created for unit testing of the AAS HTTP Client.")

    # add Submodel to AAS
    sdk_tools.add_submodel_to_aas(aas, shared_sm)

    return aas

def test_000a_create_client_by_url(client: AasHttpClient):
    base_url: str = client.base_url
    new_client = create_by_url(base_url=base_url)
    assert new_client is not None

def test_000b_create_client_by_dict(client: AasHttpClient):
    base_url: str = client.base_url

    config_dict: dict = {
        "BaseUrl": base_url
    }

    new_client = create_by_dict(configuration=config_dict)
    assert new_client is not None

def test_001a_connect(client: AasHttpClient):
    assert client is not None

def test_001b_delete_all_asset_administration_shells(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    result = client.shells.get_all_asset_administration_shells()
    assert result is not None
    shells = result.get("result", [])

    for shell in shells:
        shell_id = shell.get("id", "")

        if client.encoded_ids:
            shell_id = encoder.encode_base_64(shell_id)

        if shell_id:
            delete_result = client.shells.delete_asset_administration_shell_by_id(shell_id)
            assert delete_result

    shells_result = client.shells.get_all_asset_administration_shells()

    if shells_result is None:
        raise RuntimeError("Failed to retrieve shells after deletion.")

    shells = shells_result.get("result", [])
    assert len(shells) == 0

def test_001c_delete_all_submodels(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    result = client.submodels.get_all_submodels()
    assert result is not None
    submodels = result.get("result", [])

    for submodel in submodels:
        submodel_id = submodel.get("id", "")

        if client.encoded_ids:
            submodel_id = encoder.encode_base_64(submodel_id)

        if submodel_id:
            delete_result = client.submodels.delete_submodel_by_id(submodel_id)
            assert delete_result

    submodels_result = client.submodels.get_all_submodels()

    if submodels_result is None:
        raise RuntimeError("Failed to retrieve submodels after deletion.")

    submodels = submodels_result.get("result", [])
    assert len(submodels) == 0

def test_002_get_all_asset_administration_shells(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    result = client.shells.get_all_asset_administration_shells()
    assert result is not None
    shells = result.get("result", [])
    assert len(shells) == 0

def test_003_post_asset_administration_shell(client: AasHttpClient, shared_aas: model.AssetAdministrationShell):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    aas_data_string = json.dumps(shared_aas, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    aas_data = json.loads(aas_data_string)
    result = client.shells.post_asset_administration_shell(aas_data)

    assert result is not None
    assert result.get("idShort", "") == shared_aas.id_short
    assert result.get("id", "") == SHELL_ID

    get_result = client.shells.get_all_asset_administration_shells()
    assert get_result is not None
    shells = get_result.get("result", [])
    assert len(shells) == 1
    assert shells[0].get("idShort", "") == shared_aas.id_short
    assert shells[0].get("id", "") == SHELL_ID
    submodels = shells[0].get("submodels", [])
    assert len(submodels) == 1
    submodel: dict = submodels[0]
    assert len(submodel.get("keys", [])) == 1
    assert submodel.get("keys", [])[0].get("value", "") == SM_ID

def test_004a_get_asset_administration_shell_by_id(client: AasHttpClient, shared_aas: model.AssetAdministrationShell):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.get_asset_administration_shell_by_id(shell_id)

    assert result is not None
    assert result.get("idShort", "") == shared_aas.id_short
    assert result.get("id", "") == SHELL_ID

def test_004b_get_asset_administration_shell_by_id(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    result = client.shells.get_asset_administration_shell_by_id("non_existent_id")

    assert result is None

def test_005a_put_asset_administration_shell_by_id(client: AasHttpClient, shared_aas: model.AssetAdministrationShell):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    assert shared_aas.asset_information is not None
    assert shared_aas.asset_information.global_asset_id is not None

    aas = model.AssetAdministrationShell(id_=shared_aas.asset_information.global_asset_id, asset_information=shared_aas.asset_information)
    aas.id_short = shared_aas.id_short

    description_text = "Put description for unit tests"
    aas.description = model.MultiLanguageTextType({"en": description_text})
    aas.submodel = shared_aas.submodel  # Keep existing submodels

    aas_data_string = json.dumps(aas, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    aas_data = json.loads(aas_data_string)

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.put_asset_administration_shell_by_id(shell_id, aas_data)
    assert result

    get_result = client.shells.get_asset_administration_shell_by_id(shell_id)
    assert get_result
    assert get_result.get("idShort", "") == shared_aas.id_short
    assert get_result.get("id", "") == SHELL_ID
    # description must have changed
    assert get_result.get("description", {})[0].get("text", "") == description_text
    assert shared_aas.description is not None
    assert get_result.get("description", {})[0].get("text", "") != shared_aas.description.get("en", "")
    # submodels must be retained
    assert len(get_result.get("submodels", [])) == len(shared_aas.submodel)

    # The display name must be empty
    # NOTE: currently not working in dotnet
    # assert len(get_result.get("displayName", {})) == 0

    # restore to its original state
    sm_data_string = json.dumps(shared_aas, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)
    client.shells.put_asset_administration_shell_by_id(shell_id, sm_data)  # Restore original submodel

def test_005b_put_asset_administration_shell_by_id(client: AasHttpClient, shared_aas: model.AssetAdministrationShell):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    # put with other ID
    id_short = "put_short_id"
    identifier = f"fluid40/{id_short}"
    asset_info = model_builder.create_base_asset_information(identifier)
    assert asset_info.global_asset_id == identifier

    aas = model.AssetAdministrationShell(id_=asset_info.global_asset_id, asset_information=asset_info)
    aas.id_short = id_short

    description_text = {"en": "Updated description for unit tests"}
    aas.description = model.MultiLanguageTextType(description_text)

    aas_data_string = json.dumps(aas, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    aas_data = json.loads(aas_data_string)

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: Python server crashes by this test
        result = False
    else:
        result = client.shells.put_asset_administration_shell_by_id(shell_id, aas_data)

    assert not result

    get_result = client.shells.get_asset_administration_shell_by_id(shell_id)
    assert get_result is not None
    assert get_result.get("description", {})[0].get("text", "") != description_text
    assert shared_aas.description is not None
    assert get_result.get("description", {})[0].get("text", "") == shared_aas.description.get("en", "")

def test_006_get_asset_administration_shell_by_id_reference_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.get_asset_administration_shell_by_id_reference_aas_repository(shell_id)

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS:
        # NOTE: Basyx java server do not provide this endpoint
        assert result is None
    else:
        assert result is not None
        keys = result.get("keys", [])
        assert len(keys) == 1
        assert keys[0].get("value", "") == SHELL_ID

def test_007_get_submodel_by_id_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID
    sm_id = SM_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.shells.get_submodel_by_id_aas_repository(shell_id, sm_id)

    assert result is None

def test_008_get_all_submodels(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    result = client.submodels.get_all_submodels()
    assert result is not None
    submodels = result.get("result", [])
    assert len(submodels) == 0

def test_009a_post_submodel(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm_data_string = json.dumps(shared_sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)

    result = client.submodels.post_submodel(sm_data)

    assert result is not None
    result_id_short = result.get("idShort", "")
    assert result_id_short == shared_sm.id_short

    get_result = client.submodels.get_all_submodels()
    assert get_result is not None
    submodels = get_result.get("result", [])
    assert len(submodels) == 1
    assert submodels[0].get("idShort", "") == shared_sm.id_short

def test_009b_post_submodel(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm_template_file = Path(f"./tests/test_data/aimc.json").resolve()

    with Path.open(sm_template_file, "r", encoding="utf-8") as f:
        sm_data = json.load(f)

    result = client.submodels.post_submodel(sm_data)

    assert result is not None
    result_id = result.get("id", "")
    assert result_id == AIMC_SM_ID

    get_result = client.submodels.get_all_submodels()
    assert get_result is not None
    submodels = get_result.get("result", [])
    assert len(submodels) == 2

def test_010_get_submodel_by_id_aas_repository(client: AasHttpClient, shared_sm: model.Submodel):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID
    sm_id = SM_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.shells.get_submodel_by_id_aas_repository(shell_id, sm_id)

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS:
        # NOTE: Basyx java server do not provide this endpoint
        assert result is None
    else:
        assert result is not None
        result_id_short = result.get("idShort", "")
        assert result_id_short == shared_sm.id_short

def test_011a_get_submodel_by_id(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.get_submodel_by_id(sm_id)

    assert result is not None
    result_id_short = result.get("idShort", "")
    assert result_id_short == shared_sm.id_short

def test_011b_get_submodel_by_id(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    result = client.submodels.get_submodel_by_id("non_existent_id")

    assert result is None

def test_011c_get_submodel_by_id(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm_id = AIMC_SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = client.submodels.get_submodel_by_id(sm_id)

    assert result is not None
    result_id = result.get("id", "")
    assert result_id == AIMC_SM_ID

def test_011d_get_submodel_by_id(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm_id = AIMC_SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = client.submodels.get_submodel_by_id(sm_id, level="core")

    assert result is not None
    result_id = result.get("id", "")
    assert result_id == AIMC_SM_ID
    #assert "submodelElements" not in result

def test_012_patch_submodel_by_id(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodel API is not available in this client")

    sm = model.Submodel(shared_sm.id_short)
    sm.id_short = shared_sm.id_short

    description_text = "Patched description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})

    sm_data_string = json.dumps(sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.patch_submodel_by_id(sm_id, sm_data)

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS or parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: Basyx java and python server do not provide this endpoint
        assert not result
    else:
        assert result is True

        get_result = client.submodels.get_submodel_by_id(sm_id)
        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sm.id_short
        assert get_result.get("id", "") == SM_ID
        # Only the description may change in patch.
        assert get_result.get("description", {})[0].get("text", "") == description_text
        assert shared_sm.description is not None
        assert shared_sm.display_name is not None
        assert get_result.get("description", {})[0].get("text", "") != shared_sm.description.get("en", "")
        # The display name must remain the same.
        assert get_result.get("displayName", {})[0].get("text", "") == shared_sm.display_name.get("en", "")

def test_013_put_submodel_by_id_aas_repository(client: AasHttpClient, shared_sm: model.Submodel):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    sm = model.Submodel(SM_ID)
    sm.id_short = shared_sm.id_short

    description_text = "Put via shell description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})

    sm_data_string = json.dumps(sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)

    shell_id = SHELL_ID
    sm_id = SM_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.shells.put_submodel_by_id_aas_repository(shell_id, sm_id, sm_data)

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS:
        # NOTE: Basyx java server do not provide this endpoint
        assert not result
    else:
        assert result

        get_result = client.shells.get_submodel_by_id_aas_repository(shell_id, sm_id)
        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sm.id_short
        assert get_result.get("id", "") == SM_ID
        # description must have changed
        assert get_result.get("description", {})[0].get("text", "") == description_text
        assert shared_sm.description is not None
        assert get_result.get("description", {})[0].get("text", "") != shared_sm.description.get("en", "")
        assert len(get_result.get("displayName", {})) == 0

    # restore to its original state
    sm_data_string = json.dumps(shared_sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)
    client.shells.put_submodel_by_id_aas_repository(shell_id, sm_id, sm_data)  # Restore original submodel

def test_014_put_submodels_by_id(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm = model.Submodel(SM_ID)
    sm.id_short = shared_sm.id_short

    description_text = "Put description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})

    sm_data_string = json.dumps(sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.put_submodels_by_id(sm_id, sm_data)

    assert result is True

    get_result = client.submodels.get_submodel_by_id(sm_id)
    assert get_result is not None
    assert get_result.get("idShort", "") == shared_sm.id_short
    assert get_result.get("id", "") == SM_ID
    # description must have changed
    assert get_result.get("description", {})[0].get("text", "") == description_text
    assert shared_sm.description is not None
    assert get_result.get("description", {})[0].get("text", "") != shared_sm.description.get("en", "")
    assert len(get_result.get("displayName", {})) == 0

    # restore to its original state
    sm_data_string = json.dumps(shared_sm, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sm_data = json.loads(sm_data_string)
    client.submodels.put_submodels_by_id(SM_ID, sm_data)  # Restore original submodel

def test_015_get_all_submodel_elements_submodel_repository(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_elements = client.submodels.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.get("result", [])) == 0

def test_016a_post_submodel_element_submodel_repo(client: AasHttpClient, shared_sme_string: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sme_data_string = json.dumps(shared_sme_string, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sme_data = json.loads(sme_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.post_submodel_element_submodel_repo(sm_id, sme_data)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_string.id_short
    assert shared_sme_string.description is not None
    assert shared_sme_string.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_string.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_string.display_name.get("en", "")
    assert result.get("value", "") == shared_sme_string.value

    get_result = client.submodels.get_all_submodel_elements_submodel_repository(sm_id)
    assert get_result is not None

    assert len(get_result.get("result", [])) == 1

def test_016b_post_submodel_element_submodel_repo(client: AasHttpClient, shared_sme_bool: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sme_data_string = json.dumps(shared_sme_bool, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sme_data = json.loads(sme_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.post_submodel_element_submodel_repo(sm_id, sme_data)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_bool.id_short
    assert shared_sme_bool.description is not None
    assert shared_sme_bool.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_bool.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_bool.display_name.get("en", "")
    assert json.loads(result.get("value", "").lower()) == shared_sme_bool.value

    get_result = client.submodels.get_all_submodel_elements_submodel_repository(sm_id)
    assert get_result is not None

    assert len(get_result.get("result", [])) == 2

def test_016c_post_submodel_element_submodel_repo(client: AasHttpClient, shared_sme_int: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sme_data_string = json.dumps(shared_sme_int, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sme_data = json.loads(sme_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.post_submodel_element_submodel_repo(sm_id, sme_data)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_int.id_short
    assert shared_sme_int.description is not None
    assert shared_sme_int.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_int.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_int.display_name.get("en", "")
    assert int(result.get("value", "")) == shared_sme_int.value

    get_result = client.submodels.get_all_submodel_elements_submodel_repository(sm_id)
    assert get_result is not None

    assert len(get_result.get("result", [])) == 3

def test_016d_post_submodel_element_submodel_repo(client: AasHttpClient, shared_sme_float: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sme_data_string = json.dumps(shared_sme_float, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sme_data = json.loads(sme_data_string)

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.post_submodel_element_submodel_repo(sm_id, sme_data)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_float.id_short
    assert shared_sme_float.description is not None
    assert shared_sme_float.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_float.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_float.display_name.get("en", "")
    assert float(result.get("value", "")) == shared_sme_float.value

    get_result = client.submodels.get_all_submodel_elements_submodel_repository(sm_id)
    assert get_result is not None

    assert len(get_result.get("result", [])) == 4

def test_017a_get_submodel_element_by_path_submodel_repo(client: AasHttpClient, shared_sme_string: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_string.id_short
    assert shared_sme_string.description is not None
    assert shared_sme_string.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_string.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_string.display_name.get("en", "")
    assert result.get("value", "") == shared_sme_string.value

def test_017b_get_submodel_element_by_path_submodel_repo(client: AasHttpClient, shared_sme_bool: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_bool.id_short
    assert shared_sme_bool.description is not None
    assert shared_sme_bool.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_bool.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_bool.display_name.get("en", "")
    assert json.loads(result.get("value", "").lower()) == shared_sme_bool.value

def test_017c_get_submodel_element_by_path_submodel_repo(client: AasHttpClient, shared_sme_int: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_int.id_short
    assert shared_sme_int.description is not None
    assert shared_sme_int.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_int.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_int.display_name.get("en", "")
    assert int(result.get("value", "")) == shared_sme_int.value

def test_017d_get_submodel_element_by_path_submodel_repo(client: AasHttpClient, shared_sme_float: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

    assert result is not None
    assert result.get("idShort", "") == shared_sme_float.id_short
    assert shared_sme_float.description is not None
    assert shared_sme_float.display_name is not None
    assert result.get("description", {})[0].get("text", "") == shared_sme_float.description.get("en", "")
    assert result.get("displayName", {})[0].get("text", "") == shared_sme_float.display_name.get("en", "")
    assert float(result.get("value", "")) == shared_sme_float.value

def test_018a_patch_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_string: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    new_value = "Patched String Value"

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    assert submodel_element is not None

    old_value = submodel_element.get("value", "")

    result = client.submodels.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_string.id_short, new_value)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sme_string.id_short
        assert get_result.get("value", "") == new_value
        assert shared_sme_string.description is not None
        assert shared_sme_string.display_name is not None
        assert get_result.get("description", {})[0].get("text", "") == shared_sme_string.description.get("en", "")
        assert get_result.get("displayName", {})[0].get("text", "") == shared_sme_string.display_name.get("en", "")
        assert get_result.get("value", "") != old_value

def test_018b_patch_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_bool: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    new_value = "false"

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)
    assert submodel_element is not None

    old_value = submodel_element.get("value", "")

    result = client.submodels.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_bool.id_short, new_value)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)

        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sme_bool.id_short
        assert json.loads(get_result.get("value", "").lower()) == json.loads(new_value)
        assert shared_sme_bool.description is not None
        assert shared_sme_bool.display_name is not None
        assert get_result.get("description", {})[0].get("text", "") == shared_sme_bool.description.get("en", "")
        assert get_result.get("displayName", {})[0].get("text", "") == shared_sme_bool.display_name.get("en", "")
        assert get_result.get("value", "").lower() != old_value.lower()

def test_018c_patch_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_int: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    new_value = "263"

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    assert submodel_element is not None

    old_value = submodel_element.get("value", "")

    result = client.submodels.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_int.id_short, new_value)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)

        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sme_int.id_short
        assert int(get_result.get("value", "")) == int(new_value)
        assert shared_sme_int.description is not None
        assert shared_sme_int.display_name is not None
        assert get_result.get("description", {})[0].get("text", "") == shared_sme_int.description.get("en", "")
        assert get_result.get("displayName", {})[0].get("text", "") == shared_sme_int.display_name.get("en", "")
        assert int(get_result.get("value", "")) != int(old_value)

def test_018d_patch_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_float: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    new_value = "262.1"

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)
    assert submodel_element is not None

    old_value = submodel_element.get("value", "")

    result = client.submodels.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_float.id_short, new_value)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

        assert get_result is not None
        assert get_result.get("idShort", "") == shared_sme_float.id_short
        assert float(get_result.get("value", "")) == float(new_value)
        assert shared_sme_float.description is not None
        assert shared_sme_float.display_name is not None
        assert get_result.get("description", {})[0].get("text", "") == shared_sme_float.description.get("en", "")
        assert get_result.get("displayName", {})[0].get("text", "") == shared_sme_float.display_name.get("en", "")
        assert float(get_result.get("value", "")) != float(old_value)

def test_019a_post_submodel_element_by_path_submodel_repo(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    submodel_element_list = model.SubmodelElementList(id_short="sme_list_1", type_value_list_element=model.Property, value_type_list_element=model.datatypes.String)
    submodel_element_list_dict = sdk_tools.convert_to_dict(submodel_element_list)
    assert submodel_element_list_dict is not None

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    post_list_element_result = client.submodels.post_submodel_element_submodel_repo(sm_id, submodel_element_list_dict)

    assert post_list_element_result is not None

    property = model_builder.create_base_submodel_element_property(None, model.datatypes.String, "Value in List")
    property_dict = sdk_tools.convert_to_dict(property)
    assert property_dict is not None
    assert "idShort" not in property_dict

    result = client.submodels.post_submodel_element_by_path_submodel_repo(sm_id, submodel_element_list.id_short, property_dict)

    assert result is not None
    assert "idShort" not in result  # idShort was deleted

    submodel = client.submodels.get_submodel_by_id(sm_id)

    assert submodel is not None
    elements = submodel.get("submodelElements", [])
    assert len(elements) == 5  # 4 previous properties + 1 list
    assert elements[4].get("idShort", "") == submodel_element_list.id_short
    list_elements = elements[4].get("value", [])
    assert len(list_elements) == 1
    assert list_elements[0].get("idShort", "") == ""
    assert list_elements[0].get("value", "") == property.value


def test_019b_post_submodel_element_by_path_submodel_repo(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    submodel_element_collection = model.SubmodelElementCollection(id_short="sme_collection_1")
    submodel_element_collection_dict = sdk_tools.convert_to_dict(submodel_element_collection)
    assert submodel_element_collection_dict is not None

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    first_result = client.submodels.post_submodel_element_submodel_repo(sm_id, submodel_element_collection_dict)

    assert first_result is not None

    property = model_builder.create_base_submodel_element_property("sme_property_in_collection", model.datatypes.String, "Value in List")
    property_dict = sdk_tools.convert_to_dict(property)
    assert property_dict is not None

    result = client.submodels.post_submodel_element_by_path_submodel_repo(sm_id, submodel_element_collection.id_short, property_dict)

    assert result is not None
    assert result["idShort"] == property.id_short

    submodel = client.submodels.get_submodel_by_id(sm_id)

    assert submodel is not None
    elements = submodel.get("submodelElements", [])
    assert len(elements) == 6
    assert elements[5].get("idShort", "") == submodel_element_collection.id_short
    list_elements = elements[5].get("value", [])
    assert len(list_elements) == 1
    assert list_elements[0].get("idShort", "") == property.id_short
    assert list_elements[0].get("value", "") == property.value

    base_url: str = client.base_url
    new_client = create_by_url(base_url=base_url)
    assert new_client is not None
    assert new_client.submodels is not None

    sm = new_client.submodels.get_submodel_by_id(AIMC_SM_ID)
    assert sm is None

    decoded_id = encoder.encode_base_64(AIMC_SM_ID)
    decoded_sm = new_client.submodels.get_submodel_by_id(decoded_id)
    assert decoded_sm is not None
    assert decoded_sm.get("id", "") == AIMC_SM_ID

def test_020b_encoded_ids(client: AasHttpClient):
    base_url: str = client.base_url
    new_client = create_by_url(base_url=base_url)
    assert new_client is not None
    assert new_client.shells is not None

    sm = new_client.shells.get_asset_administration_shell_by_id(SHELL_ID)
    assert sm is None

    decoded_id = encoder.encode_base_64(SHELL_ID)
    decoded_sm = new_client.shells.get_asset_administration_shell_by_id(decoded_id)
    assert decoded_sm is not None
    assert decoded_sm.get("id", "") == SHELL_ID

def test_021_post_file_by_path_submodel_repo(client: AasHttpClient):
    if client.experimental is None:
        pytest.skip("Experimental API is not available in this client")

    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS or parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    file_sme = model.File("file_sme", content_type="application/pdf")
    file_post_result = client.submodels.post_submodel_element_submodel_repo(sm_id, sdk_tools.convert_to_dict(file_sme))
    assert file_post_result is not None

    filename = "https.pdf"
    file = Path(f"./tests/test_data/{filename}").resolve()
    result = client.experimental.post_file_by_path_submodel_repo(sm_id, file_sme.id_short, file)
    assert result is True

    result_sme = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, file_sme.id_short)

    assert result_sme is not None
    assert result_sme.get("idShort", "") == file_sme.id_short
    assert result_sme.get("contentType", "") == file_sme.content_type
    assert "value" in result_sme
    assert result_sme.get("value", "") == f"/{filename}"

def test_022_get_file_content_by_path_submodel_repo(client: AasHttpClient):
    if client.experimental is None:
        pytest.skip("Experimental API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS or parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.experimental.get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert result is not None
    assert len(result) > 0
    assert result.startswith(b"%PDF-1.7")

def test_023_put_file_content_by_path_submodel_repo(client: AasHttpClient):
    if client.experimental is None:
        pytest.skip("Experimental API is not available in this client")

    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS or parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    filename = "aimc.json"
    file = Path(f"./tests/test_data/{filename}").resolve()
    result = client.experimental.put_file_by_path_submodel_repo(sm_id, "file_sme", file)
    assert result is True

    get_result = client.experimental.get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert get_result is not None
    assert len(get_result) > 0
    assert get_result.startswith(b"{\n")

    result_sme = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, "file_sme")
    assert result_sme is not None
    assert "value" in result_sme
    assert result_sme.get("value", "") == f"/{filename}"

def test_024_delete_file_content_by_path_submodel_repo(client: AasHttpClient):
    if client.experimental is None:
        pytest.skip("Experimental API is not available in this client")

    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in JAVA_SERVER_PORTS or parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        return

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.experimental.delete_file_by_path_submodel_repo(sm_id, "file_sme")
    assert result is True

    get_result = client.experimental.get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert get_result is None

    result_sme = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, "file_sme")
    assert result_sme is not None
    assert "value" in result_sme
    assert result_sme.get("value", "") == None

def test_025_get_thumbnail_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.get_thumbnail_aas_repository(shell_id)
    assert result is None

def test_026_put_thumbnail_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    filename = "Pen_Machine.png"
    file = Path(f"./tests/test_data/{filename}").resolve()

    result = client.shells.put_thumbnail_aas_repository(shell_id, file.name, file)
    assert result is True

def test_027_get_thumbnail_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.get_thumbnail_aas_repository(shell_id)
    assert result is not None
    assert len(result) > 0
    assert result.startswith(b"\x89PNG\r\n\x1a\n")

def test_028_delete_thumbnail_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        return

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.delete_thumbnail_aas_repository(shell_id)
    assert result is True

    get_result = client.shells.get_thumbnail_aas_repository(shell_id)
    assert get_result is None

def test_029_get_all_submodel_references_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.get_all_submodel_references_aas_repository(shell_id)
    assert result is not None
    references = result.get("result", [])
    assert len(references) == 1

def test_030_post_submodel_reference_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    id = "temp_sm_id"
    id_short = "TempSM"
    temp_sml_ref = model.ModelReference.from_referable(model_builder.create_base_submodel(identifier=id, id_short=id_short))

    result = client.shells.post_submodel_reference_aas_repository(shell_id, sdk_tools.convert_to_dict(temp_sml_ref))

    assert result is not None
    assert len(result.get("keys", [])) > 0
    key: dict = result.get("keys", [])[0]
    assert key.get("value", "") == id
    assert key.get("type", "") == "Submodel"

    check_result = client.shells.get_all_submodel_references_aas_repository(shell_id)
    assert check_result is not None
    check_references = check_result.get("result", [])
    assert len(check_references) == 2

def test_031_delete_submodel_reference_by_id_aas_repository(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Submodels API is not available in this client")

    shell_id = SHELL_ID
    sm_id = "temp_sm_id"

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(sm_id)

    result = client.shells.delete_submodel_reference_by_id_aas_repository(shell_id, sm_id)

    assert result is True

    get_result = client.shells.get_all_submodel_references_aas_repository(shell_id)
    assert get_result is not None
    references = get_result.get("result", [])
    assert len(references) == 1

def test_032_put_submodel_element_by_path_submodel_repo(client: AasHttpClient, shared_sme_string: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    assert get_result is not None
    old_value = get_result.get("value", "")

    new_value = "New Value via PUT"

    sme_data_string = json.dumps(shared_sme_string, cls=basyx.aas.adapter.json.AASToJsonEncoder)
    sme_data = json.loads(sme_data_string)
    sme_data["value"] = new_value

    result = client.submodels.put_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short, sme_data)

    assert result is True

    get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

    assert get_result is not None
    assert get_result.get("idShort", "") == shared_sme_string.id_short
    assert get_result.get("value", "") == new_value
    assert get_result.get("value", "") != old_value
    assert shared_sme_string.description is not None
    assert shared_sme_string.display_name is not None
    assert get_result.get("description", {})[0].get("text", "") == shared_sme_string.description.get("en", "")
    assert get_result.get("displayName", {})[0].get("text", "") == shared_sme_string.display_name.get("en", "")

    # restore original value
    sme_data["value"] = "Sample String Value"
    result = client.submodels.put_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short, sme_data)

def test_033a_get_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_string: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    value = client.submodels.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_string.id_short)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    sm_data = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    assert sm_data is not None
    assert value == sm_data.get("value", "")

def test_033b_get_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_int: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    value = client.submodels.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_int.id_short)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    sm_data = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    assert sm_data is not None
    assert int(value) == int(sm_data.get("value", ""))

def test_033c_get_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_float: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    value = client.submodels.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_float.id_short)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    sm_data = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)
    assert sm_data is not None
    assert float(value) == float(sm_data.get("value", ""))

def test_033d_get_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_bool: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    value = client.submodels.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_bool.id_short)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    sm_data = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)
    assert sm_data is not None
    #assert bool(value) == bool(sm_data.get("value", ""))

# def test_033e_get_submodel_element_by_path_value_only_submodel_repo(client: AasHttpClient, shared_sme_collection: model.SubmodelElementCollection):
#     if client.submodels is None:
#         pytest.skip("Submodels API is not available in this client")

#     post_result = client.submodels.post_submodel_element_submodel_repo(SM_ID, sdk_tools.convert_to_dict(shared_sme_collection))
#     assert post_result is not None

#     sm_id = SM_ID

#     if client.encoded_ids:
#         sm_id = encoder.encode_base_64(SM_ID)

#     value = client.submodels.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_collection.id_short)

#     parsed = urlparse(client.base_url)
#     if parsed.port in PYTHON_SERVER_PORTS:
#         # NOTE: python server do not provide this endpoint
#         assert value is None
#         return

#     assert value is not None

#     sm_data = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_collection.id_short)
#     assert sm_data is not None
#     assert value == sm_data.get("value", "")
#     assert bool(value) == bool(sm_data.get("value", ""))

def test_034_get_submodel_by_id_value_only(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    response = client.submodels.get_submodel_by_id_value_only(sm_id)

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert response is None
        return
    elif parsed.port in DOTNET_SERVER_PORTS:
        assert response is not None
        value = response[shared_sm.id_short]
    else:
        assert response is not None
        value = response

    assert value is not None
    assert len(value) > 5
    assert "sme_property_int" in value
    assert int(value["sme_property_int"]) == 263
    assert "sme_property_string" in value
    assert value["sme_property_string"] == "Sample String Value"
    assert "sme_property_float" in value
    assert float(value["sme_property_float"]) == 262.1

def test_035_patch_submodel_by_id_value_only(client: AasHttpClient, shared_sm: model.Submodel, shared_sme_string: model.Property, shared_sme_int: model.Property, shared_sme_float: model.Property):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    value_dict = {
        shared_sme_string.id_short: shared_sme_string.value,
        shared_sme_int.id_short: str(shared_sme_int.value),
        shared_sme_float.id_short: str(shared_sme_float.value)
    }

    # patch_dict = {shared_sm.id: value_dict}

    patch_dict = value_dict

    parsed = urlparse(client.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        return

    if parsed.port in JAVA_SERVER_PORTS:
        # NOTE: java server endpoint seems to work not correctly
        return

    elif parsed.port in DOTNET_SERVER_PORTS:
        patch_dict = value_dict

    result = client.submodels.patch_submodel_by_id_value_only(sm_id, patch_dict)

    assert result is True

    string_prop_dict = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    int_prop_dict = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    float_prop_dict = client.submodels.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

    assert string_prop_dict is not None
    assert int_prop_dict is not None
    assert float_prop_dict is not None

    assert string_prop_dict.get("value", "") == shared_sme_string.value
    assert int(int_prop_dict.get("value", "")) == int(shared_sme_int.value)
    assert float(float_prop_dict.get("value", "")) == float(shared_sme_float.value)


def test_036_get_submodel_by_id_metadata(client: AasHttpClient, shared_sm: model.Submodel):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    metadata = client.submodels.get_submodel_by_id_metadata(sm_id)
    assert metadata is not None

    submodel = client.submodels.get_submodel_by_id(sm_id)
    assert submodel is not None

    assert metadata.get("id", "") == submodel.get("id", "")
    assert metadata.get("idShort", "") == submodel.get("idShort", "")
    assert metadata.get("description", {})[0].get("text", "") == submodel.get("description", {})[0].get("text", "")
    if "displayName" in submodel and submodel.get("displayName", []):
        assert metadata.get("displayName", {})[0].get("text", "") == submodel.get("displayName", {})[0].get("text", "")
    assert "submodelElements" not in metadata

def test_098_delete_asset_administration_shell_by_id(client: AasHttpClient):
    if client.shells is None:
        pytest.skip("Shells API is not available in this client")

    shell_id = SHELL_ID

    if client.encoded_ids:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = client.shells.delete_asset_administration_shell_by_id(shell_id)

    assert result is True

    get_result = client.shells.get_all_asset_administration_shells()
    assert get_result is not None
    shells = get_result.get("result", [])
    assert len(shells) == 0

def test_099a_delete_submodel_by_id(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(SM_ID)

    result = client.submodels.delete_submodel_by_id(sm_id)

    assert result is True

    get_result = client.submodels.get_all_submodels()
    assert get_result is not None
    submodels = get_result.get("result", [])
    assert len(submodels) == 1

def test_099b_delete_submodel_by_id(client: AasHttpClient):
    if client.submodels is None:
        pytest.skip("Submodels API is not available in this client")

    sm_id = AIMC_SM_ID

    if client.encoded_ids:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = client.submodels.delete_submodel_by_id(sm_id)

    assert result is True

    get_result = client.submodels.get_all_submodels()
    assert get_result is not None
    submodels = get_result.get("result", [])
    assert len(submodels) == 0
